import logging
import uuid
from datetime import datetime

import pytest

from pytest_instrument.backport_configurable_stacklevel import patch_logger
from pytest_instrument.logging_helpers import setup_log_file_handler, InstLogger


def pytest_addoption(parser):
    parser.addoption(
        "--instrument",
        action="store",
        default=None,
        type=lambda s: [item for item in s.split(",")],
        help="enable pytest-instrument output; comma-separated list; expected valued: json, log",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "instrument: pytest-instrument mark")


def pytest_sessionstart(session):
    session_id = str(uuid.uuid4())

    log_handlers = []
    if session.config.getoption("instrument") is not None:
        current_timestamp = datetime.now().strftime("%Y%m%dT%H%M%S")
        if "json" in session.config.getoption("instrument"):
            base_filename = f"{current_timestamp}_{session_id[:8]}"
            log_handler_json = setup_log_file_handler(base_filename, "json")
            log_handlers.append(log_handler_json)
        if "log" in session.config.getoption("instrument"):
            base_filename = f"{current_timestamp}_{session_id[:8]}"
            log_handler_plain = setup_log_file_handler(base_filename, "log")
            log_handlers.append(log_handler_plain)

            record = {
                "name": "instr.report",
                "node_id": "",
                "levelname": logging.getLevelName(logging.INFO),
                "levelno": logging.INFO,
                "msg": f"session id: {session_id}",
            }

            log_record = logging.makeLogRecord(record)
            log_handler_plain.emit(log_record)
    else:
        log_handlers.append(logging.NullHandler())

    logging.setLoggerClass(patch_logger(InstLogger))
    logger = logging.getLogger("instr.log")
    logger.setLevel("DEBUG")
    for handler in log_handlers:
        logger.addHandler(handler)

    logger.session_id = session_id

    session.config.instrument = {
        "session_id": session_id,
        "logger": logger,
        "logfile_handler": log_handlers,
    }


def pytest_sessionfinish(session, exitstatus):
    for handler in session.config.instrument["logfile_handler"]:
        handler.close()
        session.config.instrument["logger"].removeHandler(handler)


def pytest_addhooks(pluginmanager):
    from pytest_instrument import hooks

    pluginmanager.add_hookspecs(hooks)


def pytest_runtest_setup(item):
    try:
        labels = list([_.args for _ in item.iter_markers("instrument")][0])
    except IndexError:
        labels = []

    item.config.hook.pytest_instrument_labels(config=item.config, labels=labels)

    try:
        tags = [_.kwargs for _ in item.iter_markers("instrument")][0]
    except IndexError:
        tags = {}

    item.config.hook.pytest_instrument_tags(config=item.config, tags=tags)

    labels = None if not labels else labels
    tags = None if not tags else tags
    labels_and_tags = {"labels": labels, "tags": tags}
    item.user_properties.append(("instrument", labels_and_tags))

    fixtures = item.fixturenames.copy()
    item.config.hook.pytest_instrument_fixtures(fixtures=fixtures)
    fixtures = None if not fixtures else fixtures
    item.user_properties.append(("fixtures", fixtures))

    item.config.instrument["logger"].node_id = item._nodeid

    for logger in [
        logging.getLogger(name)
        for name in logging.root.manager.loggerDict
        if "instr.log" in name
    ]:
        logger.node_id = item._nodeid


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    item.user_properties.append((call.when, {"start": call.start, "stop": call.stop}))

    outcome = yield

    report = outcome.get_result()
    _log_report(report, item.config)


def _log_report(report, config):
    if config.getoption("instrument") is not None:
        labels_and_tags = {}
        for prop in (
            prop for prop in report.user_properties if prop[0] == "instrument"
        ):
            labels_and_tags = prop[1]

        timestamps = {}
        for prop in (prop for prop in report.user_properties if prop[0] == report.when):
            timestamps = prop[1]

        fixtures = []
        for prop in (prop for prop in report.user_properties if prop[0] == "fixtures"):
            fixtures = prop[1]

        record = {
            "name": "instr.report",
            # ToDo: set level based on passed, skipped, failed
            "levelname": logging.getLevelName(logging.INFO),
            "levelno": logging.INFO,
            "msg": f"{report.nodeid} {report.when} {report.outcome}",
            "session_id": config.instrument["session_id"],
            "node_id": report.nodeid,
            "when": report.when,
            "outcome": report.outcome,
            "start": str(timestamps["start"]),
            "stop": str(timestamps["stop"]),
            "duration": f"{report.duration:.12f}",
            "labels": labels_and_tags.get("labels", None),
            "tags": labels_and_tags.get("tags", None),
            "fixtures": fixtures,
        }

        log_record = logging.makeLogRecord(record)
        for handler in config.instrument["logfile_handler"]:
            handler.emit(log_record)

    else:
        pass



def pytest_instrument_labels(config, labels):
    """Called in pytest_runtest_setup(item) after collecting labels from @pytest.mark.instrument"""


def pytest_instrument_tags(config, tags):
    """Called in pytest_runtest_setup(item) after collecting tags from @pytest.mark.instrument"""


def pytest_instrument_fixtures(fixtures):
    """Called in pytest_runtest_setup(item) after collecting fixturenames from item"""



import logging
import os
import uuid
from datetime import datetime

from pythonjsonlogger import jsonlogger


class InstLogger(logging.Logger):
    def __init__(self, name):
        super().__init__(name)
        self.session_id = None
        self.node_id = None

    def getChild(self, suffix):
        child_logger = super().getChild(suffix)
        child_logger.session_id = self.session_id
        child_logger.node_id = self.node_id
        return child_logger

    def debug(self, msg, extra=None, *args, **kwargs):
        extra = {} if extra is None else extra
        y = {"session_id": self.session_id, "node_id": self.node_id}
        super().debug(msg, extra={**extra, **y}, stacklevel=2, *args, **kwargs)

    def info(self, msg, extra=None, *args, **kwargs):
        extra = {} if extra is None else extra
        y = {"session_id": self.session_id, "node_id": self.node_id}
        super().info(msg, extra={**extra, **y}, stacklevel=2, *args, **kwargs)

    def warning(self, msg, extra=None, *args, **kwargs):
        extra = {} if extra is None else extra
        y = {"session_id": self.session_id, "node_id": self.node_id}
        super().warning(msg, extra={**extra, **y}, stacklevel=2, *args, **kwargs)

    def critical(self, msg, extra=None, *args, **kwargs):
        extra = {} if extra is None else extra
        y = {"session_id": self.session_id, "node_id": self.node_id}
        super().critical(msg, extra={**extra, **y}, stacklevel=2, *args, **kwargs)

    def error(self, msg, extra=None, *args, **kwargs):
        extra = {} if extra is None else extra
        y = {"session_id": self.session_id, "node_id": self.node_id}
        super().error(msg, extra={**extra, **y}, stacklevel=2, *args, **kwargs)


def setup_log_file_handler(filename, output_format):
    try:
        os.mkdir("./artifacts", mode=0o777)
    except FileExistsError:
        pass
    log_file = f"./artifacts/{filename}.{output_format}"

    if output_format == "json":
        return json_logfile_handler(log_file)
    elif output_format == "log":
        return log_logfile_handler(log_file)


def json_logfile_handler(logfile):
    formatter = CustomJsonFormatter(
        "%(timestamp) %(level) %(name) %(message) %(filename)s %(funcName)s %(lineno)d"
    )
    log_handler = logging.FileHandler(logfile)
    log_handler.setFormatter(formatter)
    return log_handler


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)
        if not log_record.get("timestamp"):
            log_record["timestamp"] = datetime.fromtimestamp(record.created).strftime(
                "%Y-%m-%d %H:%M:%S.%f"
            )
        if log_record.get("level"):
            log_record["level"] = log_record["level"].lower()
        else:
            log_record["level"] = record.levelname.lower()
        if not log_record.get("record_id"):
            log_record["record_id"] = str(uuid.uuid4())


def log_logfile_handler(logfile):
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(node_id)s - %(message)s"
    )
    log_handler = logging.FileHandler(logfile)
    log_handler.setFormatter(formatter)
    return log_handler



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import codecs
from setuptools import setup


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding="utf-8").read()


setup(
    name="pytest-instrument",
    version="0.3.0",
    author="Joep Schuurkes",
    author_email="j19sch@gmail.com",
    maintainer="Joep Schuurkes",
    maintainer_email="j19sch@gmail.com",
    license="MIT",
    url="https://github.com/j19sch/pytest-instrument",
    description="pytest plugin to instrument tests",
    long_description=read("README.rst"),
    packages=["pytest_instrument"],
    python_requires=">=3.6",
    install_requires=["pytest>=5.1.0", "python-json-logger>=0.1.11"],
    tests_require=["jsonschema>=3.1.1"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Testing",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={"pytest11": ["metrics = pytest_instrument.plugin"]},
)



# source: https://github.com/tonysyu/logquacious/blob/master/logquacious/backport_configurable_stacklevel.py

"""
Backport of configurable stacklevel for logging added in Python 3.8.

See https://github.com/python/cpython/pull/7424
"""
import io
import logging
import os
import sys
import traceback
from contextlib import contextmanager


__all__ = ["PatchedLoggerMixin", "patch_logger"]


class PatchedLoggerMixin(object):
    """Mixin adding `temp_monkey_patched_logger` that allows stacklevel kwarg.

    Classes that include this mixin have a `temp_monkey_patched_logger`
    context manager that allows the use of the `stacklevel` keyword argument
    from Python 3.8.

    Classes using this mixin must have a `logging.Logger` instance as an
    attribute of the class. By default, this is assumed to be named `logger`,
    but you can override the `logger_attribute` class attribute with the
    name of a different attribute.
    """

    #: Name of logger instance on the class inheriting this mixin.
    logger_attribute = "logger"

    def __init__(self, *args, **kwargs):
        super(PatchedLoggerMixin, self).__init__(*args, **kwargs)
        self._patched_logger_class = None

    def _get_logger(self):
        if not hasattr(self, self.logger_attribute):
            msg = (
                "Subclass of PatchedLoggerMixin must define `{}` attribute "
                "or override `logger_attribute`".format(self.logger_attribute)
            )
            raise AttributeError(msg)
        return getattr(self, self.logger_attribute)

    @contextmanager
    def temp_monkey_patched_logger(self):
        """Temporarily monkey patch logger to allow overriding log records.

        The monkey patching is reset so that the behavior change is limited
        to the scope of this logger.
        """
        logger = self._get_logger()
        original_logger_class = logger.__class__

        # Cache patched logger class if not already defined.
        if self._patched_logger_class is None:
            self._patched_logger_class = patch_logger(logger.__class__)

        logger.__class__ = self._patched_logger_class
        try:
            yield
        finally:
            logger.__class__ = original_logger_class


def patch_logger(logger_class):
    """Return logger class patched with stacklevel keyword argument."""
    if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
        return logger_class
    return type(
        "ConfigurableStacklevelLogger",
        (ConfigurableStacklevelLoggerMixin, logger_class),
        {},
    )


class ConfigurableStacklevelLoggerMixin(object):
    """Mixin for adding `stacklevel` keyword argument for logging methods.

    This mixin can be used to monkey patch `logging.Logger` to include the
    `stacklevel` keyword argument that will be available in Python 3.8.

    See https://github.com/python/cpython/pull/7424
    """

    def findCaller(self, stack_info=False, stacklevel=1):  # pragma: no cover
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = logging.currentframe()
        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        orig_f = f
        while f and stacklevel > 1:
            f = f.f_back
            stacklevel -= 1
        if not f:
            f = orig_f
        rv = "(unknown file)", 0, "(unknown function)", None
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == logging._srcfile:
                f = f.f_back
                continue
            sinfo = None
            if stack_info:
                sio = io.StringIO()
                sio.write("Stack (most recent call last):\n")
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == "\n":
                    sinfo = sinfo[:-1]
                sio.close()
            rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
            break
        return rv

    def _log(
        self,
        level,
        msg,
        args,
        exc_info=None,
        extra=None,
        stack_info=False,
        stacklevel=1,
    ):  # pragma: no cover
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.
        """
        sinfo = None
        if logging._srcfile:
            # IronPython doesn't track Python frames, so findCaller raises an
            # exception on some versions of IronPython. We trap it here so that
            # IronPython can use logging.
            try:
                fn, lno, func, sinfo = self.findCaller(stack_info, stacklevel)
            except ValueError:
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

        if sys.version_info.major >= 3:
            record = self.makeRecord(
                self.name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
            )
        else:
            record = self.makeRecord(
                self.name, level, fn, lno, msg, args, exc_info, func, extra
            )
        self.handle(record)
